# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from typing import Any, Dict

from babel.dates import LOCALTZ
from flask import Flask
from werkzeug.datastructures import ImmutableDict

from abilian.web.action import Endpoint

default_config = dict(Flask.default_config)  # type: Dict[str, Any]
default_config.update(
    PRIVATE_SITE=False,
    TEMPLATE_DEBUG=False,
    CSRF_ENABLED=True,
    BABEL_ACCEPT_LANGUAGES=["en"],
    DEFAULT_COUNTRY=None,
    PLUGINS=(),
    ADMIN_PANELS=(
        "abilian.web.admin.panels.dashboard.DashboardPanel",
        "abilian.web.admin.panels.audit.AuditPanel",
        "abilian.web.admin.panels.login_sessions.LoginSessionsPanel",
        "abilian.web.admin.panels.settings.SettingsPanel",
        "abilian.web.admin.panels.users.UsersPanel",
        "abilian.web.admin.panels.groups.GroupsPanel",
        "abilian.web.admin.panels.sysinfo.SysinfoPanel",
        "abilian.web.admin.panels.impersonate.ImpersonatePanel",
        "abilian.services.vocabularies.admin.VocabularyPanel",
        "abilian.web.tags.admin.TagPanel",
    ),

    CELERYD_MAX_TASKS_PER_CHILD=1000,
    CELERY_ACCEPT_CONTENT=["pickle", "json", "msgpack", "yaml"],
    CELERY_TIMEZONE=LOCALTZ,

    SENTRY_SDK_URL="https://browser.sentry-cdn.com/4.5.3/bundle.min.js",
    # SENTRY_USER_ATTRS=("email", "first_name", "last_name"),
    # SENTRY_INSTALL_CLIENT_JS=True,  # also install client JS
    # SENTRY_JS_VERSION="3.26.4",
    # TODO: remove, not needed for recent sentry-js
    # SENTRY_JS_PLUGINS=("console", "jquery", "native", "require"),

    SESSION_COOKIE_NAME=None,

    SQLALCHEMY_POOL_RECYCLE=1800,  # 30min. default value in flask_sa is None
    SQLALCHEMY_TRACK_MODIFICATIONS=False,

    LOGO_URL=Endpoint("abilian_static", filename="img/logo-abilian-32x32.png"),
    ABILIAN_UPSTREAM_INFO_ENABLED=False,  # upstream info extension
    TRACKING_CODE_SNIPPET="",  # tracking code to insert before </body>
    MAIL_ADDRESS_TAG_CHAR=None,
)
default_config = ImmutableDict(default_config)



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import sys
import tempfile
from pathlib import Path
from warnings import warn

from magic import Magic, os
from pytest import fixture, mark

from abilian.services.conversion.handlers import HAS_LIBREOFFICE, HAS_PDFTOTEXT

mime_sniffer = Magic(mime=True)
encoding_sniffer = Magic(mime_encoding=True)

# FIXME: tests that rely on OOo are disabled until we fix stability issues.


@fixture
def converter():
    from abilian.services.conversion import converter as c

    cache_dir = tempfile.mkdtemp(suffix="unittest")
    tmp_dir = tempfile.mkdtemp(suffix="unittest")
    c.init_work_dirs(cache_dir, tmp_dir)
    yield c

    c.clear()


def read_file(fn, mode="rb"):
    return (Path(__file__).parent / "dummy_files" / fn).open(mode).read()


# To text
@mark.skipif(not HAS_PDFTOTEXT, reason="requires poppler or poppler-util")
def test_pdf_to_text(converter):
    blob = read_file("onepage.pdf")
    text = converter.to_text("", blob, "application/pdf")
    assert text


@mark.skipif(not HAS_LIBREOFFICE, reason="requires libreoffice")
def test_word_to_text(converter):
    blob = read_file("test.doc")
    text = converter.to_text("", blob, "application/msword")
    assert text


@mark.skipif(not HAS_LIBREOFFICE, reason="requires libreoffice")
def test_wordx_to_text(converter):
    blob = read_file("test.docx")
    text = converter.to_text("", blob, "application/msword")
    assert text


# def test_excel_to_text(converter):
#     blob = read_file("test.xls")
#     text = converter.to_text("", blob, "application/excel")


# To PDF
@mark.skipif(not HAS_LIBREOFFICE, reason="requires libreoffice")
def test_odt_to_pdf(converter):
    blob = read_file("test.odt")
    pdf = converter.to_pdf("", blob, "application/vnd.oasis.opendocument.text")
    assert "application/pdf" == mime_sniffer.from_buffer(pdf)


@mark.skipif(not HAS_LIBREOFFICE, reason="requires libreoffice")
def test_word_to_pdf(converter):
    blob = read_file("test.doc")
    pdf = converter.to_pdf("", blob, "application/msword")
    assert "application/pdf" == mime_sniffer.from_buffer(pdf)


@mark.skipif(
    sys.platform.startswith("linux"), reason="maybe bug due to ImagMagik settings"
)
def test_image_to_pdf(converter):
    blob = read_file("picture.jpg")
    pdf = converter.to_pdf("", blob, "image/jpeg")
    assert "application/pdf" == mime_sniffer.from_buffer(pdf)


# To images
@mark.skipif(not HAS_PDFTOTEXT, reason="requires poppler or poppler-util")
def test_pdf_to_images(converter):
    if not os.popen("which pdftoppm").read().strip():
        warn("pdftoppm not found, skipping test")
        return
    blob = read_file("onepage.pdf")
    image = converter.to_image("", blob, "application/pdf", 0)
    assert "image/jpeg" == mime_sniffer.from_buffer(image)


@mark.skipif(not HAS_PDFTOTEXT, reason="requires poppler or poppler-util")
def test_word_to_images(converter):
    blob = read_file("test.doc")
    image = converter.to_image("", blob, "application/msword", 0)
    assert "image/jpeg" == mime_sniffer.from_buffer(image)



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from datetime import datetime

from pytest import fixture
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from abilian.core.entities import Entity
from abilian.core.models.base import AUDITABLE, NOT_SEARCHABLE, SEARCHABLE, \
    Info
from abilian.core.models.subjects import User

from .dummy import DummyContact


@fixture
def session():
    engine = create_engine("sqlite:///:memory:", echo=False)
    session_class = sessionmaker(bind=engine)
    session = session_class()

    # flask-sqlalchemy as listeners looking for this
    session._model_changes = {}

    DummyContact.metadata.create_all(engine)
    return session


def test():
    contact = DummyContact(first_name="John")
    assert contact.creator is None
    assert contact.owner is None

    user = User()
    contact.owner = user
    contact.creator = user


def test_auto_slug_property(session):
    obj = DummyContact(name="a b c")
    session.add(obj)
    session.flush()
    assert obj.auto_slug == "a-b-c"

    obj.name = "C'est l'été !"
    assert obj.auto_slug == "c-est-l-ete"

    # with a special space character
    obj.name = "a_b\u205fc"  # U+205F: MEDIUM MATHEMATICAL SPACE
    assert obj.auto_slug == "a-b-c"

    # with non-ascii translatable chars, like EN DASH U+2013 (–) and EM DASH
    # U+2014 (—). Standard separator is \u002d (\x2d) "-" HYPHEN-MINUS.
    # this test may fails depending on how  Unicode normalization + char
    # substitution is done (order matters).
    obj.name = "a\u2013b\u2014c"  # u'a–b—c'
    slug = obj.auto_slug
    assert slug == "a-b-c"
    assert "\u2013" not in slug
    assert "\u002d" in slug


def test_updated_at(session):
    contact = DummyContact()
    session.add(contact)
    session.commit()
    assert isinstance(contact.updated_at, datetime)

    updated = contact.updated_at
    contact.first_name = "John"
    session.commit()
    assert isinstance(contact.updated_at, datetime)
    assert contact.updated_at > updated


def test_auto_slug(session):
    contact1 = DummyContact(name="Pacôme Hégésippe Adélard Ladislas")
    session.add(contact1)
    session.flush()
    assert contact1.slug == "pacome-hegesippe-adelard-ladislas"

    # test when name is None
    contact2 = DummyContact()
    session.add(contact2)
    session.flush()
    expected = "dummycontact-{}".format(contact2.id)
    assert contact2.slug == expected

    # test numbering if slug already exists:
    contact3 = DummyContact(name="Pacôme Hégésippe Adélard Ladislas")
    session.add(contact3)
    session.flush()
    assert contact3.slug == "pacome-hegesippe-adelard-ladislas-1"


def test_polymorphic_update_timestamp(session):
    contact = DummyContact(name="Pacôme Hégésippe Adélard Ladislas")
    session.add(contact)
    session.flush()

    updated_at = contact.updated_at
    assert updated_at
    contact.email = "p@example.com"
    session.flush()
    assert contact.updated_at > updated_at


def test_meta(session):
    e = DummyContact(name="test")
    e.meta["key"] = "value"
    e.meta["number"] = 42
    session.add(e)
    session.flush()
    e_id = e.id
    session.expunge(e)
    del e
    e = session.query(DummyContact).get(e_id)
    assert e.meta["key"] == "value"
    assert e.meta["number"] == 42


def test_entity_type():
    class MyType(Entity):
        pass

    expected = __name__ + ".MyType"
    assert MyType.entity_type == expected
    assert MyType._object_type() == expected

    class Fixed(Entity):
        entity_type = "some.fixed.module.fixed_type"

    assert Fixed.entity_type == "some.fixed.module.fixed_type"
    assert Fixed._object_type() == "some.fixed.module.fixed_type"

    class OtherBase(Entity):
        ENTITY_TYPE_BASE = "some.module"

    assert OtherBase.entity_type == "some.module.OtherBase"
    assert OtherBase._object_type() == "some.module.OtherBase"

    # test when ENTITY_TYPE_BASE is in ancestors
    class Base(object):
        ENTITY_TYPE_BASE = "from.ancestor"

    class InheritedBase(Base, Entity):
        pass

    assert InheritedBase.entity_type == "from.ancestor.InheritedBase"
    assert InheritedBase._object_type() == "from.ancestor.InheritedBase"


def test_info():
    info = SEARCHABLE
    assert info["searchable"]

    info = NOT_SEARCHABLE
    assert not info["searchable"]

    info = SEARCHABLE + AUDITABLE
    assert info["searchable"]
    assert info["auditable"]
    assert isinstance(info, Info)

    info = SEARCHABLE | AUDITABLE
    assert info["searchable"]
    assert info["auditable"]
    assert isinstance(info, Info)



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import socket
from collections import OrderedDict, namedtuple
from pathlib import Path

import redis
import six
import sqlalchemy as sa
import sqlalchemy.dialects
import sqlalchemy.exc
from flask import (
    current_app,
    flash,
    redirect,
    render_template,
    request,
    session,
    url_for,
)
from six import text_type
from typing import Dict, Text

from abilian.core.commands import config as cmd_config
from abilian.core.extensions import csrf, db
from abilian.core.models.subjects import User
from abilian.core.util import unwrap
from abilian.services import get_service
from abilian.services.security import Admin, Anonymous
from abilian.web.blueprints import Blueprint

logger = logging.getLogger(__name__)
setup = Blueprint(
    "setup", __name__, allowed_roles=Anonymous, template_folder="templates"
)

# list supported dialects and detect unavailable ones due to missing dbapi
# module ('psycopg2' missing for example)

_dialects = OrderedDict((("sqlite", "SQLite (for demo)"), ("postgresql", "PostgreSQL")))

_dialects_unavailable = OrderedDict()  # type: Dict[Text, Text]

for dialect, _label in _dialects.items():
    d = sa.dialects.registry.load(dialect)
    try:
        d.dbapi()
    except ImportError as e:
        _dialects_unavailable[dialect] = str(e)

# enumerate steps for left column progress
Step = namedtuple("SetupStep", ("name", "endpoint", "title", "description"))

_setup_steps = (
    Step("db", "step_db", "Setup Database", "Setup basic database connection"),
    Step("redis", "step_redis", "Setup Redis", "Redis connection"),
    Step(
        "site_info",
        "step_site_info",
        "Basic site informations",
        "Site name, admin email...",
    ),
    Step("admin_account", "step_admin_account", "Admin account", None),
    Step("finalize", "finalize", "Finalize", None),
)


@setup.before_request
def step_progress():
    session.permanent = False  # limit lifetime to browser session
    current_step = next_step = prev_step = None

    for step in _setup_steps:
        if current_step is not None:
            next_step = step
            break
        endpoint = "{}.{}".format(setup.name, step.endpoint)
        if endpoint == request.endpoint:
            current_step = step
        else:
            prev_step = step

    request.setup_step_progress = {
        "current_step": current_step,
        "prev_step": prev_step,
        "next_step": next_step,
    }


@setup.context_processor
def common_context():
    ctx = {
        "setup_steps": _setup_steps,
        "dialects": _dialects,
        "dialects_unavailable": _dialects_unavailable,
        "validated_steps": session_get("validated", ()),
    }
    ctx.update(request.setup_step_progress)
    return ctx


# session helpers
def session_set(key, data):
    session.setdefault("abilian_setup", {})[key] = data


def session_get(key, default=None):
    return session.setdefault("abilian_setup", {}).get(key, default)


def session_clear():
    if "abilian_setup" in session:
        del session["abilian_setup"]


def step_validated(step_endpoint, valid=True):
    validated = set(session_get("validated", ()))
    if valid:
        validated.add(step_endpoint)
    else:
        if step_endpoint in validated:
            validated.remove(step_endpoint)

    session_set("validated", tuple(validated))


#
# DB Setup
#
@csrf.exempt
@setup.route("/", methods=["GET", "POST"])
def step_db():
    if request.method == "POST":
        return step_db_validate()
    return step_db_form()


def step_db_form():
    return render_template("setupwizard/step_db.html", data=session_get("db", {}))


def step_db_validate():
    form = request.form
    dialect = form["dialect"]
    username = form.get("username", "").strip()
    password = form.get("password", "").strip()
    host = form.get("host", "").strip()
    port = form.get("port", "").strip()
    database = form.get("database", "").strip()

    # build db_uri
    db_uri = "{}://".format(dialect)

    if dialect != "sqlite":
        if username:
            db_uri += username
            # check password only if we have a username
            if password:
                db_uri += ":" + password

        # FIXME: it is an error to have a username and no host,
        # SA will interpret username:password as host:port
        if host:
            db_uri += "@" + host
            port = port
            if port:
                db_uri += ":" + port

    db_uri += "/"

    if dialect == "sqlite" and (not database or database == ":memory:"):
        database = text_type(Path(current_app.instance_path) / "data" / "sqlite.db")

    db_uri += database

    # store in session
    db_params = {
        "uri": db_uri,
        "dialect": dialect,
        "username": username,
        "password": password,
        "host": host,
        "port": port,
        "database": database,
    }
    session_set("db", db_params)

    # test connection
    engine = sa.create_engine(db_uri)
    error = None
    try:
        conn = engine.connect()
    except sa.exc.OperationalError as e:
        error = str(e)
    else:
        conn.close()
        del conn

    engine.dispose()

    # FIXME: at this point we have not tested the connection is valid, i.e, we
    # have CREATE, SELECT, UPDATE, DELETE rights on the database

    if error:
        flash(error, "error")
        return step_db_form()

    step_validated("db")
    next_step = request.setup_step_progress["next_step"]
    endpoint = "{}.{}".format(setup.name, next_step.endpoint)
    return redirect(url_for(endpoint))


#
# Redis
#
@csrf.exempt
@setup.route("/redis", methods=["GET", "POST"])
def step_redis():
    if request.method == "POST":
        return step_redis_validate()
    return step_redis_form()


def step_redis_form():
    return render_template("setupwizard/step_redis.html", data=session_get("redis", {}))


def step_redis_validate():
    form = request.form
    data = {
        "host": form.get("host", "localhost").strip(),
        "port": form.get("port", "").strip() or "6379",
        "db": form.get("db", "").strip() or "1",
    }

    for k in ("port", "db"):
        try:
            data[k] = int(data[k])
        except ValueError:
            pass

    data["uri"] = "redis://{host}:{port}/{db}".format(**data)
    session_set("redis", data)
    error = None

    try:
        r = redis.StrictRedis(host=data["host"], port=data["port"], db=data["db"])
    except Exception as e:
        error = "Connection error, check parameters"
        raise e

    try:
        r.info()
    except redis.exceptions.InvalidResponse:
        error = (
            "Connection error: doesn't look like it's a redis server. "
            "Verify host and port are those of your redis server."
        )
    except redis.exceptions.ResponseError as e:
        error = "Redis server response: {}".format(e)
    except redis.exceptions.RedisError as e:
        error = "Unknown redis error ({})".format(e)

    if error:
        flash(error, "error")
        return step_redis_form()

    step_validated("redis")
    next_step = request.setup_step_progress["next_step"]
    endpoint = "{}.{}".format(setup.name, next_step.endpoint)
    return redirect(url_for(endpoint))


#
# Site info
#
@csrf.exempt
@setup.route("/site_info", methods=["GET", "POST"])
def step_site_info():
    if request.method == "POST":
        return step_site_info_validate()
    return step_site_info_form()


def get_possible_hostnames():
    hostname = socket.gethostname()
    fqdn = socket.getfqdn()
    names = {"localhost": ["127.0.0.1"]}

    for name in (hostname, fqdn):
        try:
            name, aliases, ips = socket.gethostbyname_ex(name)
        except socket.error:
            continue
        names.setdefault(name, []).extend(ips)
        for a in aliases:
            names.setdefault(a, []).extend(ips)

    return sorted(
        "{} ({})".format(name, ", ".join(sorted(set(ips))))
        for name, ips in six.iteritems(names)
    )


def step_site_info_form():
    cfg = current_app.config
    default_data = {
        "sitename": cfg.get("SITE_NAME", "") or "",
        "mailsender": cfg.get("MAIL_SENDER", "") or "",
    }
    return render_template(
        "setupwizard/step_site_info.html",
        data=session_get("site_info", default_data),
        suggested_hosts=get_possible_hostnames(),
    )


def step_site_info_validate():
    form = request.form
    data = {
        "sitename": form.get("sitename", "").strip(),
        "mailsender": form.get("mailsender", "").strip(),
        "server_mode": form.get("server_mode", "").strip(),
    }

    session_set("site_info", data)
    step_validated("site_info")
    next_step = request.setup_step_progress["next_step"]
    endpoint = "{}.{}".format(setup.name, next_step.endpoint)
    return redirect(url_for(endpoint))


#
# Admin account creation
#
@csrf.exempt
@setup.route("/admin_account", methods=["GET", "POST"])
def step_admin_account():
    if request.method == "POST":
        return step_admin_account_validate()
    return step_admin_account_form()


def step_admin_account_form():
    return render_template(
        "setupwizard/step_admin_account.html", data=session_get("admin_account", {})
    )


def step_admin_account_validate():
    form = request.form
    password = form.get("password", "").strip()
    confirm_password = form.get("confirm_password", "").strip()
    admin_user = {
        "email": form.get("email", "").strip(),
        "name": form.get("name", "").strip(),
        "firstname": form.get("firstname", "").strip(),
        "password": password,
        "confirm_password": confirm_password,
    }
    session_set("admin_account", admin_user)

    if password != confirm_password:
        flash("Password fields don't match", "error")
        return step_admin_account_form()

    step_validated("admin_account")
    next_step = request.setup_step_progress["next_step"]
    endpoint = "{}.{}".format(setup.name, next_step.endpoint)
    return redirect(url_for(endpoint))


#
# Finalize
#
@csrf.exempt
@setup.route("/finalize", methods=["GET", "POST"])
def finalize():
    validated = session_get("validated")
    assert all(s.name in validated for s in _setup_steps[:-1])

    if request.method == "POST":
        return finalize_validate()
    return finalize_form()


def finalize_form():
    file_location = Path(current_app.instance_path) / "config.py"
    return render_template("setupwizard/finalize.html", file_location=file_location)


def finalize_validate():
    config_file = Path(current_app.instance_path) / "config.py"
    logging_file = Path(current_app.instance_path) / "logging.yml"

    assert not config_file.exists()
    config = cmd_config.DefaultConfig(logging_file="logging.yml")
    config.SQLALCHEMY_DATABASE_URI = session_get("db")["uri"]

    redis_uri = session_get("redis")["uri"]
    config.REDIS_URI = redis_uri
    config.BROKER_URL = redis_uri
    config.CELERY_RESULT_BACKEND = redis_uri

    d = session_get("site_info")
    config.SITE_NAME = d["sitename"]
    config.MAIL_SENDER = d["mailsender"]

    is_production = d["server_mode"] == "production"
    config.PRODUCTION = is_production
    config.DEBUG = not is_production
    config.DEBUG_TB_ENABLED = config.DEBUG
    config.CELERY_ALWAYS_EAGER = not is_production

    cmd_config.write_config(config_file, config)
    cmd_config.maybe_write_logging(logging_file)

    admin_account = session_get("admin_account")
    # create a new app that will be configured with new config,
    # to create database and admin_user
    setup_app = unwrap(current_app)
    app = setup_app.__class__(
        setup_app.import_name,
        static_url_path=setup_app.static_url_path,
        static_folder=setup_app.static_folder,
        template_folder=setup_app.template_folder,
        instance_path=setup_app.instance_path,
    )
    with app.test_request_context("/setup/finalize"):
        app.create_db()
        db_session = db.session()
        admin = User(
            email=admin_account["email"],
            password=admin_account["password"],
            last_name=admin_account["name"],
            first_name=admin_account["firstname"],
            can_login=True,
        )
        db_session.add(admin)
        security = get_service("security")
        security.grant_role(admin, Admin)
        db_session.commit()

    session_clear()

    return render_template(
        "setupwizard/done.html", config_file=config_file, logging_file=logging_file
    )



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import sqlalchemy as sa
from pytest import fixture

from abilian.core.entities import Entity


class IndexedContact(Entity):
    # default is 'test_service.IndexedContact'
    entity_type = "abilian.services.indexing.IndexedContact"
    name = sa.Column(sa.UnicodeText)


@fixture
def svc(app):
    svc = app.services["indexing"]
    with app.app_context():
        svc.start()
        yield svc


def test_app_state(app, svc):
    state = svc.app_state
    assert IndexedContact in state.indexed_classes
    assert IndexedContact.entity_type in svc.adapted
    assert IndexedContact.entity_type in state.indexed_fqcn


def test_index_only_after_final_commit(app, session, svc):
    contact = IndexedContact(name="John Doe")

    state = svc.app_state

    session.begin(nested=True)
    assert state.to_update == []

    session.add(contact)
    # no commit: model is in wait queue
    session.flush()
    assert state.to_update == [("new", contact)]

    # commit but in a sub transaction: model still in wait queue
    session.commit()
    assert state.to_update == [("new", contact)]

    # 'final' commit: models sent for indexing update
    session.commit()
    assert state.to_update == []


def test_clear(app, svc):
    # just check no exception happens
    svc.clear()

    # check no double stop (would raise AssertionError from service base)
    svc.start()
    svc.stop()
    svc.clear()



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

from flask import redirect, request, url_for
from flask_login import login_user

from abilian.core.models.subjects import User
from abilian.web.admin.panel import AdminPanel


class ImpersonatePanel(AdminPanel):
    id = "impersonate"
    label = "Impersonate"
    icon = "user"

    def get(self):
        user_email = request.args.get("email")
        if not user_email:
            return "UI not done yet. Add '?email=...' at the end of this URL for now."

        user = User.query.filter(User.email == user_email).one()
        login_user(user)
        return redirect(url_for("main.home"))



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['abilian',
 'abilian.core',
 'abilian.core.commands',
 'abilian.core.extensions',
 'abilian.core.models',
 'abilian.core.models.tests',
 'abilian.core.tests',
 'abilian.services',
 'abilian.services.activity',
 'abilian.services.antivirus',
 'abilian.services.audit',
 'abilian.services.auth',
 'abilian.services.conversion',
 'abilian.services.image',
 'abilian.services.image.tests',
 'abilian.services.indexing',
 'abilian.services.preferences',
 'abilian.services.repository',
 'abilian.services.security',
 'abilian.services.settings',
 'abilian.services.tagging',
 'abilian.services.viewtracker',
 'abilian.services.vocabularies',
 'abilian.testing',
 'abilian.web',
 'abilian.web.admin',
 'abilian.web.admin.panels',
 'abilian.web.admin.panels.groups',
 'abilian.web.admin.panels.users',
 'abilian.web.assets',
 'abilian.web.attachments',
 'abilian.web.comments',
 'abilian.web.coreviews',
 'abilian.web.forms',
 'abilian.web.preferences',
 'abilian.web.search',
 'abilian.web.setupwizard',
 'abilian.web.tags',
 'abilian.web.tests',
 'abilian.web.tools',
 'abilian.web.uploads',
 'abilian.web.views']

package_data = \
{'': ['*'],
 'abilian': ['translations/*',
             'translations/de_DE/LC_MESSAGES/messages.po',
             'translations/en/LC_MESSAGES/messages.po',
             'translations/es/LC_MESSAGES/messages.po',
             'translations/es_ES/LC_MESSAGES/messages.po',
             'translations/fr/LC_MESSAGES/messages.po',
             'translations/fr_FR/LC_MESSAGES/messages.po',
             'translations/id_ID/LC_MESSAGES/messages.po',
             'translations/it_IT/LC_MESSAGES/messages.po',
             'translations/nl/LC_MESSAGES/messages.po',
             'translations/nl_NL/LC_MESSAGES/messages.po',
             'translations/pt_BR/LC_MESSAGES/messages.po',
             'translations/pt_PT/LC_MESSAGES/messages.po',
             'translations/tr_TR/LC_MESSAGES/messages.po',
             'translations/zh_CN/LC_MESSAGES/messages.po',
             'translations/zh_TW/LC_MESSAGES/messages.po'],
 'abilian.core.commands': ['templates/*'],
 'abilian.services': ['sandbox/*'],
 'abilian.services.auth': ['templates/login/*', 'templates/login/email/*'],
 'abilian.services.conversion': ['dummy_files/*'],
 'abilian.services.indexing': ['tests/*'],
 'abilian.services.vocabularies': ['templates/admin/*',
                                   'templates/admin/macros/*'],
 'abilian.web': ['resources/bootbox/*',
                 'resources/bootstrap-datepicker/css/*',
                 'resources/bootstrap-datepicker/js/*',
                 'resources/bootstrap-datepicker/js/locales/*',
                 'resources/bootstrap-datepicker/less/*',
                 'resources/bootstrap-switch/*',
                 'resources/bootstrap-switch/less/bootstrap3/*',
                 'resources/bootstrap-timepicker/css/*',
                 'resources/bootstrap-timepicker/js/*',
                 'resources/bootstrap-timepicker/less/*',
                 'resources/bootstrap/fonts/*',
                 'resources/bootstrap/js/*',
                 'resources/bootstrap/less/*',
                 'resources/bootstrap/less/mixins/*',
                 'resources/ckeditor/*',
                 'resources/ckeditor/adapters/*',
                 'resources/ckeditor/lang/*',
                 'resources/ckeditor/plugins/*',
                 'resources/ckeditor/plugins/a11yhelp/dialogs/*',
                 'resources/ckeditor/plugins/a11yhelp/dialogs/lang/*',
                 'resources/ckeditor/plugins/about/dialogs/*',
                 'resources/ckeditor/plugins/about/dialogs/hidpi/*',
                 'resources/ckeditor/plugins/autolink/*',
                 'resources/ckeditor/plugins/bootstrapVisibility/*',
                 'resources/ckeditor/plugins/bootstrapVisibility/lang/*',
                 'resources/ckeditor/plugins/clipboard/dialogs/*',
                 'resources/ckeditor/plugins/colordialog/dialogs/*',
                 'resources/ckeditor/plugins/dialog/*',
                 'resources/ckeditor/plugins/div/dialogs/*',
                 'resources/ckeditor/plugins/find/dialogs/*',
                 'resources/ckeditor/plugins/flash/dialogs/*',
                 'resources/ckeditor/plugins/flash/images/*',
                 'resources/ckeditor/plugins/forms/dialogs/*',
                 'resources/ckeditor/plugins/forms/images/*',
                 'resources/ckeditor/plugins/iframe/dialogs/*',
                 'resources/ckeditor/plugins/iframe/images/*',
                 'resources/ckeditor/plugins/image/dialogs/*',
                 'resources/ckeditor/plugins/image/images/*',
                 'resources/ckeditor/plugins/link/dialogs/*',
                 'resources/ckeditor/plugins/link/images/*',
                 'resources/ckeditor/plugins/link/images/hidpi/*',
                 'resources/ckeditor/plugins/liststyle/dialogs/*',
                 'resources/ckeditor/plugins/magicline/images/*',
                 'resources/ckeditor/plugins/magicline/images/hidpi/*',
                 'resources/ckeditor/plugins/pagebreak/images/*',
                 'resources/ckeditor/plugins/pastefromword/filter/*',
                 'resources/ckeditor/plugins/preview/*',
                 'resources/ckeditor/plugins/scayt/*',
                 'resources/ckeditor/plugins/scayt/dialogs/*',
                 'resources/ckeditor/plugins/showblocks/images/*',
                 'resources/ckeditor/plugins/smiley/dialogs/*',
                 'resources/ckeditor/plugins/smiley/images/*',
                 'resources/ckeditor/plugins/specialchar/dialogs/*',
                 'resources/ckeditor/plugins/specialchar/dialogs/lang/*',
                 'resources/ckeditor/plugins/table/dialogs/*',
                 'resources/ckeditor/plugins/tabletools/dialogs/*',
                 'resources/ckeditor/plugins/templates/dialogs/*',
                 'resources/ckeditor/plugins/templates/templates/*',
                 'resources/ckeditor/plugins/templates/templates/images/*',
                 'resources/ckeditor/plugins/wsc/*',
                 'resources/ckeditor/plugins/wsc/dialogs/*',
                 'resources/ckeditor/samples/*',
                 'resources/ckeditor/samples/css/*',
                 'resources/ckeditor/samples/img/*',
                 'resources/ckeditor/samples/js/*',
                 'resources/ckeditor/samples/old/*',
                 'resources/ckeditor/samples/old/assets/*',
                 'resources/ckeditor/samples/old/assets/inlineall/*',
                 'resources/ckeditor/samples/old/assets/outputxhtml/*',
                 'resources/ckeditor/samples/old/assets/uilanguages/*',
                 'resources/ckeditor/samples/old/dialog/*',
                 'resources/ckeditor/samples/old/dialog/assets/*',
                 'resources/ckeditor/samples/old/enterkey/*',
                 'resources/ckeditor/samples/old/htmlwriter/*',
                 'resources/ckeditor/samples/old/htmlwriter/assets/outputforflash/*',
                 'resources/ckeditor/samples/old/magicline/*',
                 'resources/ckeditor/samples/old/toolbar/*',
                 'resources/ckeditor/samples/old/wysiwygarea/*',
                 'resources/ckeditor/samples/toolbarconfigurator/*',
                 'resources/ckeditor/samples/toolbarconfigurator/css/*',
                 'resources/ckeditor/samples/toolbarconfigurator/font/*',
                 'resources/ckeditor/samples/toolbarconfigurator/js/*',
                 'resources/ckeditor/samples/toolbarconfigurator/lib/codemirror/*',
                 'resources/ckeditor/skins/moono/*',
                 'resources/ckeditor/skins/moono/images/*',
                 'resources/ckeditor/skins/moono/images/hidpi/*',
                 'resources/datatables/css/*',
                 'resources/datatables/images/*',
                 'resources/datatables/js/*',
                 'resources/fileapi/*',
                 'resources/fileapi/plugins/*',
                 'resources/font-awesome/css/*',
                 'resources/font-awesome/fonts/*',
                 'resources/font-awesome/less/*',
                 'resources/font-awesome/scss/*',
                 'resources/highlightjs/*',
                 'resources/img/abilian.png',
                 'resources/img/abilian.png',
                 'resources/img/abilian.png',
                 'resources/img/abilian.png',
                 'resources/img/avatar-default.png',
                 'resources/img/avatar-default.png',
                 'resources/img/avatar-default.png',
                 'resources/img/avatar-default.png',
                 'resources/img/logo-abilian-32x32.png',
                 'resources/img/logo-abilian-32x32.png',
                 'resources/img/logo-abilian-32x32.png',
                 'resources/img/logo-abilian-32x32.png',
                 'resources/img/logo-abilian.png',
                 'resources/img/logo-abilian.png',
                 'resources/img/logo-abilian.png',
                 'resources/img/logo-abilian.png',
                 'resources/jquery/js/*',
                 'resources/js/*',
                 'resources/js/widgets/*',
                 'resources/less/*',
                 'resources/nvd3/*',
                 'resources/requirejs/*',
                 'resources/select2/*',
                 'resources/typeahead/*',
                 'templates/*',
                 'templates/debug_panels/*',
                 'templates/default/*',
                 'templates/macros/*',
                 'templates/preferences/*',
                 'templates/widgets/*'],
 'abilian.web.admin': ['templates/admin/*'],
 'abilian.web.preferences': ['tests/*'],
 'abilian.web.search': ['templates/search/*'],
 'abilian.web.setupwizard': ['templates/setupwizard/*'],
 'abilian.web.tags': ['templates/admin/*'],
 'abilian.web.views': ['tests/*']}

install_requires = \
['Flask-Assets>=0.12',
 'Flask-Babel>=0.11',
 'Flask-DebugToolbar>=0.10',
 'Flask-Login>=0.4',
 'Flask-Mail>=0.9.1',
 'Flask-Migrate>=2.0',
 'Flask-Script>=2.0,<3.0',
 'Flask-Testing>=0.7',
 'Flask-WTF>=0.12,<0.13',
 'Flask>=1.0,<2.0',
 'MarkupSafe>=0.21',
 'PyYAML',
 'SQLAlchemy>=1.1,<2.0',
 'WTForms-Alchemy>=0.12',
 'WTForms-Components>=0.10',
 'Whoosh>=2.5,<3.0',
 'alembic>=0.9',
 'bcrypt',
 'bleach>=2',
 'celery>=4,<5',
 'clamd',
 'closure',
 'cssmin',
 'flask-sqlalchemy<=2.1',
 'hyperlink',
 'jsmin',
 'lxml',
 'pandas>=0.17',
 'pathlib',
 'pillow',
 'pygeoip',
 'python-dateutil>=2.4.0,<3.0.0',
 'python-deprecated',
 'python-magic',
 'redis>=2.10.1,<2.11.0',
 'sentry-sdk[flask]',
 'six',
 'sqlparse',
 'tqdm',
 'wtforms-sqlalchemy']

extras_require = \
{':python_version >= "2.7" and python_version < "2.8"': ['typing']}

setup_kwargs = {
    'name': 'abilian-core',
    'version': '0.10.40',
    'description': 'A framework for enterprise applications (CRM, ERP, collaboration...), based on Flask and SQLAlchemy',
    'long_description': 'About\n=====\n\n.. image:: https://api.travis-ci.org/abilian/abilian-core.png\n   :target: https://travis-ci.org/abilian/abilian-core\n\n.. image:: https://coveralls.io/repos/abilian/abilian-core/badge.png?branch=master\n   :target: https://coveralls.io/r/abilian/abilian-core?branch=master\n\n\nAbilian Core is an enterprise application development platform based on the `Flask micro-framework <http://flask.pocoo.org/>`_, the `SQLAlchemy ORM <http://www.sqlalchemy.org/>`_, good intentions and best practices (for some value of "best").\n\nThe full documentation is available on http://docs.abilian.com/.\n\n\nGoals & principles\n------------------\n\n- Development must be easy and fun (some some definition of "easy" and "fun", of course)\n\n- The less code (and configuration) we write, the better\n\n- Leverage existing reputable open source libraries and frameworks, such as SQLAlchemy and Flask\n\n- It must lower errors, bugs, project\'s time to deliver. It\'s intended to be a rapid application development tool\n\n- It must promote best practices in software development, specially Test-Driven Development (as advocated by the `GOOS book <http://www.amazon.com/gp/product/0321503627/>`_)\n\n\nFeatures\n--------\n\nHere\'s a short list of features that you may find appealing in Abilian:\n\nInfrastructure\n^^^^^^^^^^^^^^\n\n-  Plugin framework\n\n-  Asynchronous tasks (using `Celery <http://www.celeryproject.org/>`_)\n\n-  Security model and service\n\nDomain model and services\n^^^^^^^^^^^^^^^^^^^^^^^^^\n\n-  Persistent domain object model, based on SQLAlchemy\n\n-  Audit\n\nContent management and services\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n-  Simple file-based content repository\n\n-  Indexing service\n\n-  Document preview and transformation\n\nSocial\n^^^^^^\n\n-  Users, groups and social graph (followers)\n\n-  Activity streams\n\nUser Interface and API\n^^^^^^^^^^^^^^^^^^^^^^\n\n-  Forms (based on `WTForms <http://wtforms.simplecodes.com/>`_)\n\n-  CRUD (Create, Retrieve, Edit/Update, Remove) interface from domain\n   models\n\n-  Labels and descriptions for each field\n\n-  Various web utilities: view decorators, class-based views, Jinja2\n   filters, etc.\n\n-  A default UI based on `Bootstrap 3 <http://getbootstrap.com/>`_ and\n   several carefully selected jQuery plugins such as\n   `Select2 <http://ivaynberg.github.io/select2/>`_\n\n-  REST and AJAX API helpers\n\n-  i18n: support for multi-language via Babel, with multiple translation\n   dictionaries\n\nManagement and admin\n^^^^^^^^^^^^^^^^^^^^\n\n-  Initial settings wizard\n\n-  Admin and user settings framework\n\n-  System monitoring (using `Sentry <https://getsentry.com/welcome/>`_)\n\nCurrent status\n--------------\n\nAbilian Core is currently alpha software, in terms of API stability.\n\nIt is currently used in several applications that have been developped\nby `Abilian <http://www.abilian.com/>`_ over the last two years:\n\n-  `Abilian SBE (Social Business Engine)\n   <https://github.com/abilian/abilian-sbe/>`_ - an enterprise 2.0\n   (social collaboration) platform\n\n-  Abilian EMS (Event Management System)\n\n-  Abilian CRM (Customer / Contact / Community Relationship Management\n   System)\n\n-  Abilian Le MOOC - a MOOC prototype\n\n-  Abilian CMS - a Web CMS\n\nIn other words, Abilian Core is the foundation for a small, but growing,\nfamily of business-critical applications that our customers intend us to\nsupport in the coming years.\n\nSo while Abilian Core APIs, object model and even architecture, may (and\nmost probably will) change due to various refactorings that are expected\nas we can\'t be expected to ship perfect software on the firt release, we\nalso intend to treat it as a valuable business asset and keep\nmaintaining and improving it in the foreseeable future.\n\nRoadmap & getting involved\n--------------------------\n\nIf you need help or for general discussions about the Abilian Platform, we\nrecommend joing the `Abilian Users\n<https://groups.google.com/forum/#!forum/abilian-users>`_ forum on Google\nGroups.\n\nWe have a `roadmap on Pivotal\nTracker <https://www.pivotaltracker.com/s/projects/878951>`_ that we use\ninternally to manage our iterative delivery process.\n\nFor features and bug requests (or is it the other way around?), we\nrecommend that you use the `GitHub issue\ntracker <https://github.com/abilian/abilian-core/issues>`_.\n\n\nInstall\n=======\n\nIf you are a Python web developer (which is the primary target for this\nproject), you probably already know about:\n\n-  Python 2.7\n-  Virtualenv\n-  Poetry (https://poetry.eustace.io/)\n\nSo, after you have created and activated a virtualenv for the project,\njust run::\n\n    poetry\n\nTo use some features of the library, namely document and images\ntransformation, you will need to install the additional native packages,\nusing our operating system\'s package management tools (``dpkg``,\n``yum``, ``brew``...):\n\n-  A few image manipulation libraries (``libpng``, ``libjpeg``)\n-  The ``poppler-utils``, ``unoconv``, ``LibreOffice``, ``ImageMagick``\n   utilities\n\nLook at the ``fabfile.py`` for the exact list.\n\n\nTesting\n=======\n\nAbilian Core come with a full unit and integration testing suite. You\ncan run it with ``make test`` (once your virtualenv has been activated and\nall required dependencies have been installed, see above).\n\nAlternatively, you can use ``tox`` to run the full test suite in an\nisolated environment.\n\n\nLicence\n=======\n\nAbilian Core is licensed under the LGPL.\n\n\nCredits\n=======\n\nAbilian Core has been created by the development team at Abilian\n(currently: Stefane and Bertrand), with financial support from our\nwonderful customers, and R&D fundings from the French Government, the\nParis Region and the European Union.\n\nWe are also specially grateful to:\n\n-  `Armin Ronacher <http://lucumr.pocoo.org/>`_ for his work on Flask.\n-  `Michael Bayer <http://techspot.zzzeek.org/>`_ for his work on\n   SQLAlchemy.\n-  Everyone who has been involved with and produced open source software\n   for the Flask ecosystem (Kiran Jonnalagadda and the\n   `HasGeek <https://hasgeek.com/>`_ team, Max Countryman, Matt Wright,\n   Matt Good, Thomas Johansson, James Crasta, and probably many others).\n-  The creators of Django, Pylons, TurboGears, Pyramid and Zope, for\n   even more inspiration.\n-  The whole Python community.\n\nLinks\n=====\n\n- `Discussion list (Google Groups) <https://groups.google.com/forum/#!forum/abilian-users>`_\n- `Documentation <http://docs.abilian.com/>`_\n- `GitHub repository <https://github.com/abilian/abilian-core>`_\n- `Corporate support <http://www.abilian.com>`_\n',
    'author': 'Abilian SAS',
    'author_email': None,
    'url': 'https://github.com/abilian/abilian-core',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*',
}


setup(**setup_kwargs)



# coding=utf-8
"""Objects to schema adapters."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging
from abc import ABCMeta, abstractmethod
from operator import attrgetter

import sqlalchemy as sa
from six import add_metaclass, string_types
from whoosh.fields import TEXT

from abilian.core.extensions import db

from .schema import accent_folder

__all__ = ["SchemaAdapter", "SAAdapter"]

logger = logging.getLogger(__name__)


@add_metaclass(ABCMeta)
class SchemaAdapter(object):
    """Abstract base class for objects to schema adapter.

    The purpose of adapters is that given an object they return kwargs
    for document.
    """

    def __init__(self, model_class, schema):
        """
        :param:model_class: class of objects instances to be adapted
        :param:schema: :class:`whoosh.fields.Schema` instance
        """
        pass

    @staticmethod
    def can_adapt(obj_cls):
        """Return True if this class can adapt objects of class `obj_cls`."""
        raise NotImplementedError

    @abstractmethod
    def retrieve(self, pk, **data):
        """Returns an object instance given its identifier and optional data
        kwargs.

        :returns: None if not object instance can be returned
        """
        raise NotImplementedError

    @abstractmethod
    def get_document(self, obj):
        raise NotImplementedError


class SAAdapter(SchemaAdapter):
    """Adapts an sqlalchemy model.

    A column on the model is indexed if `"searchable"` is set to `True` in
    its `info` field (see :meth:`sqlalchemy.schema.Column.__init__`)

    Additionnaly info can also contains `index_to`. It's expected to be an
    iterable of field names or tuple of (field name, whoosh field
    instance). The default the whoosh field is a
    :class:`whoosh.fields.TEXT` instance.

    When `index_to` is used the column name will *not* be indexed on the same
    field name unless specified in `index_to`.

    If multiple field names are passed to `index_to` the values will be
    concatenated with a space.

    `index_to` is useful for::

    * index many attributes to on field names, for exemple index `title`
    and `description` to a `_full_text` field
    * define new fields on schema
    """

    @staticmethod
    def can_adapt(obj_cls):
        return issubclass(obj_cls, db.Model)

    def __init__(self, model_class, schema):
        """
        :param:model_class: a sqlalchemy model class
        :param:schema: :class:`whoosh.fields.Schema` instance
        """
        assert issubclass(model_class, db.Model)
        self.model_class = model_class
        self.indexable = getattr(model_class, "__indexable__", False)
        self.index_to = self.get_index_to(model_class)
        self.doc_attrs = {}
        if self.indexable:
            self._build_doc_attrs(model_class, schema)

    def get_index_to(self, model_class):
        result = []
        classes = model_class.mro()
        for cls in classes:
            if hasattr(cls, "__index_to__"):
                result += cls.__index_to__
        return tuple(result)

    def _build_doc_attrs(self, model_class, schema):
        mapper = sa.inspect(model_class)

        args = self.doc_attrs
        # any field not in schema will be stored here. After all field have been
        # discovered we add missing ones
        field_definitions = {}

        def setup_field(attr_name, field_name):
            field_def = False
            if not isinstance(field_name, string_types):
                field_name, field_def = field_name

            if field_name not in schema:
                if (
                    field_name not in field_definitions
                    or field_definitions[field_name] is False
                ):
                    field_definitions[field_name] = field_def

            # attrgetter offers dotted name support. Useful for attributes on
            # related objects.
            args.setdefault(field_name, {})[name] = attrgetter(name)

        # model level definitions
        for name, field_names in self.index_to:
            if isinstance(field_names, string_types):
                field_names = (field_names,)
            for field_name in field_names:
                setup_field(name, field_name)

        # per column definitions
        for col in mapper.columns:
            name = col.name
            info = col.info

            if not info.get("searchable"):
                continue

            index_to = info.get("index_to", (name,))
            if isinstance(index_to, string_types):
                index_to = (index_to,)

            for field_name in index_to:
                setup_field(name, field_name)

        # add missing fields to schema
        for field_name, field_def in field_definitions.items():
            if field_name in schema:
                continue

            if field_def is False:
                field_def = TEXT(stored=True, analyzer=accent_folder)

            logger.debug(
                "Adding field to schema:\n" "  Model: %s\n" '  Field: "%s" %s',
                model_class._object_type(),
                field_name,
                field_def,
            )
            schema.add(field_name, field_def)

    def retrieve(self, pk, _session=None, **data):
        if _session is None:
            _session = db.session()
        return _session.query(self.model_class).get(pk)

    def get_document(self, obj):
        kwargs = {}
        if not self.indexable:
            return kwargs

        # cache because the same attribute may be needed by many fields, i.e
        # "title" on "title" field and "full_text" field for example
        cached = {}
        # negative cache. Might be used especially with dotted names
        missed = set()

        for field, attrs in self.doc_attrs.items():
            values = []

            for a, getter in attrs.items():
                if a in missed:
                    continue

                if a not in cached:
                    try:
                        val = getter(obj)
                    except AttributeError:
                        missed.add(a)
                        continue
                    else:
                        cached[a] = val

                val = cached[a]
                if val is not None:
                    if isinstance(val, (list, tuple)):
                        val = " ".join(val).strip()
                    values.append(val)

            values = [v for v in values if v]

            if values:
                if len(values) == 1:
                    kwargs[field] = values[0]
                else:
                    kwargs[field] = " ".join(values)

        return kwargs
